@@ -127,7 +127,7 @@ static int write_with_timestamp(int fd, const char *data, size_t sz,
 	const char *last_byte = data + sz - 1;
 
 	while (data <= last_byte) {
-		const char *nl = memchr(data, '\n', sz);
+		const char *nl = memchr(data, '\n', last_byte + 1 - data);
 		int found_nl = (nl != NULL);
 		if (!found_nl)
 			nl = last_byte;
@@ -288,6 +288,7 @@ static int create_domain_log(struct domain *dom)
 	namepath = s;
 	strcat(namepath, "/name");
 	data = xs_read(xs, XBT_NULL, namepath, &len);
+	free(namepath);
 	if (!data)
 		return -1;
 	if (!len) {
@@ -402,9 +403,7 @@ static int domain_create_tty(struct domain *dom)
 	assert(dom->slave_fd == -1);
 	assert(dom->master_fd == -1);
 
-	cfmakeraw(&term);
-
-	if (openpty(&dom->master_fd, &dom->slave_fd, NULL, &term, NULL) < 0) {
+	if (openpty(&dom->master_fd, &dom->slave_fd, NULL, NULL, NULL) < 0) {
 		err = errno;
 		dolog(LOG_ERR, "Failed to create tty for domain-%d "
 		      "(errno = %i, %s)",
@@ -412,6 +411,22 @@ static int domain_create_tty(struct domain *dom)
 		return 0;
 	}
 
+	if (tcgetattr(dom->slave_fd, &term) < 0) {
+		err = errno;
+		dolog(LOG_ERR, "Failed to get tty attributes for domain-%d "
+			"(errno = %i, %s)",
+			dom->domid, err, strerror(err));
+		goto out;
+	}
+	cfmakeraw(&term);
+	if (tcsetattr(dom->slave_fd, TCSANOW, &term) < 0) {
+		err = errno;
+		dolog(LOG_ERR, "Failed to set tty attributes for domain-%d "
+			"(errno = %i, %s)",
+			dom->domid, err, strerror(err));
+		goto out;
+	}
+
 	if ((slave = ptsname(dom->master_fd)) == NULL) {
 		err = errno;
 		dolog(LOG_ERR, "Failed to get slave name for domain-%d "
@@ -534,9 +549,6 @@ static int domain_create_ring(struct domain *dom)
 	}
 	free(type);
 
-	if ((ring_ref == dom->ring_ref) && (remote_port == dom->remote_port))
-		goto out;
-
 	if (ring_ref != dom->ring_ref) {
 		if (dom->interface != NULL)
 			munmap(dom->interface, getpagesize());
@@ -551,6 +563,16 @@ static int domain_create_ring(struct domain *dom)
 		dom->ring_ref = ring_ref;
 	}
 
+	/* Go no further if port has not changed and we are still bound. */
+	if (remote_port == dom->remote_port) {
+		xc_evtchn_status_t status = {
+			.dom = DOMID_SELF,
+			.port = dom->local_port };
+		if ((xc_evtchn_status(xc, &status) == 0) &&
+		    (status.status == EVTCHNSTAT_interdomain))
+			goto out;
+	}
+
 	dom->local_port = -1;
 	dom->remote_port = -1;
 	if (dom->xce_handle != -1)
@@ -587,7 +609,7 @@ static int domain_create_ring(struct domain *dom)
 		}
 	}
 
-	if (log_guest)
+	if (log_guest && (dom->log_fd == -1))
 		dom->log_fd = create_domain_log(dom);
 
  out:
@@ -666,7 +688,7 @@ static struct domain *create_domain(int domid)
 	dom->buffer.capacity = 0;
 	dom->buffer.max_capacity = 0;
 	dom->event_count = 0;
-	dom->next_period = (ts.tv_sec * 1000) + (ts.tv_nsec / 1000000) + RATE_LIMIT_PERIOD;
+	dom->next_period = ((long long)ts.tv_sec * 1000) + (ts.tv_nsec / 1000000) + RATE_LIMIT_PERIOD;
 	dom->next = NULL;
 
 	dom->ring_ref = -1;
@@ -987,7 +1009,7 @@ void handle_io(void)
 
 		if (clock_gettime(CLOCK_MONOTONIC, &ts) < 0)
 			return;
-		now = (ts.tv_sec * 1000) + (ts.tv_nsec / 1000000);
+		now = ((long long)ts.tv_sec * 1000) + (ts.tv_nsec / 1000000);
 
 		/* Re-calculate any event counter allowances & unblock
 		   domains with new allowance */
