@@ -27,6 +27,11 @@
 struct xs_handle;
 typedef uint32_t xs_transaction_t;
 
+/* IMPORTANT: For details on xenstore protocol limits, see
+ * docs/misc/xenstore.txt in the Xen public source repository, and use the
+ * XENSTORE_*_MAX limit macros defined in xen/io/xs_wire.h.
+ */
+
 /* On failure, these routines set errno. */
 
 /* Connect to the xs daemon.
@@ -163,6 +168,7 @@ bool xs_is_domain_introduced(struct xs_handle *h, unsigned int domid);
 char *xs_debug_command(struct xs_handle *h, const char *cmd,
 		       void *data, unsigned int len);
 
+int xs_suspend_evtchn_port(int domid);
 #endif /* _XS_H */
 
 /*
