@@ -76,11 +76,7 @@ unsigned int acpi_cpei_phys_cpuid;
 unsigned long acpi_wakeup_address = 0;
 
 #ifdef CONFIG_IA64_GENERIC
-#ifndef XEN
 static unsigned long __init acpi_find_rsdp(void)
-#else
-unsigned long __init acpi_find_rsdp(void)
-#endif
 {
 	unsigned long rsdp_phys = 0;
 
@@ -219,6 +215,32 @@ acpi_parse_lapic_addr_ovr(struct acpi_subtable_header * header,
 	return 0;
 }
 
+#ifdef XEN
+
+#define MAX_LOCAL_SAPIC 255
+static u16 ia64_acpiid_to_sapicid[ MAX_LOCAL_SAPIC ] =
+		{[0 ... MAX_LOCAL_SAPIC - 1] = 0xffff };
+
+/* acpi id to cpu id */
+int get_cpu_id(u8 acpi_id)
+{
+	int i;
+	u16 apic_id;
+
+	apic_id = ia64_acpiid_to_sapicid[acpi_id];
+	if ( apic_id == 0xffff )
+		return -EINVAL;
+
+	for ( i = 0; i < NR_CPUS; i++ )
+	{
+		if ( apic_id == ia64_cpu_to_sapicid[i] )
+			return i;
+	}
+
+	return -1;
+}
+#endif
+
 static int __init
 acpi_parse_lsapic(struct acpi_subtable_header * header, const unsigned long end)
 {
@@ -232,6 +254,10 @@ acpi_parse_lsapic(struct acpi_subtable_header * header, const unsigned long end)
 #ifdef CONFIG_SMP
 		smp_boot_data.cpu_phys_id[available_cpus] =
 		    (lsapic->id << 8) | lsapic->eid;
+#endif
+#ifdef XEN
+        ia64_acpiid_to_sapicid[lsapic->processor_id] =
+            (lsapic->id << 8) | lsapic->eid;
 #endif
 		++available_cpus;
 	}
@@ -264,7 +290,12 @@ acpi_parse_iosapic(struct acpi_subtable_header * header, const unsigned long end
 	if (BAD_MADT_ENTRY(iosapic, end))
 		return -EINVAL;
 
+#ifndef XEN
 	return iosapic_init(iosapic->address, iosapic->global_irq_base);
+#else
+	return iosapic_init(iosapic->address, iosapic->global_irq_base,
+			    iosapic->id);
+#endif
 }
 
 static unsigned int __initdata acpi_madt_rev;
@@ -762,6 +793,10 @@ int __init acpi_boot_init(void)
 	if (acpi_table_parse(ACPI_SIG_FADT, acpi_parse_fadt))
 		printk(KERN_ERR PREFIX "Can't find FADT\n");
 
+#ifdef XEN
+	acpi_dmar_init();
+#endif
+
 #ifdef CONFIG_SMP
 	if (available_cpus == 0) {
 		printk(KERN_INFO "ACPI: Found 0 CPUS; assuming 1\n");
