@@ -242,14 +242,11 @@ static struct irqaction ipi_irqaction = {
 };
 #endif
 
-#ifdef XEN
-extern void setup_vector (unsigned int vec, struct irqaction *action);
-#endif
-
 void
 register_percpu_irq (ia64_vector vec, struct irqaction *action)
 {
 	irq_desc_t *desc;
+#ifndef XEN
 	unsigned int irq;
 
 	for (irq = 0; irq < NR_IRQS; ++irq)
@@ -258,14 +255,51 @@ register_percpu_irq (ia64_vector vec, struct irqaction *action)
 			desc->status |= IRQ_PER_CPU;
 			desc->handler = &irq_type_ia64_lsapic;
 			if (action)
-#ifdef XEN
-				setup_vector(irq, action);
-#else
 				setup_irq(irq, action);
-#endif
 		}
+#else
+	desc = irq_descp(vec);
+	desc->status |= IRQ_PER_CPU;
+	desc->handler = &irq_type_ia64_lsapic;
+	if (action)
+		setup_vector(vec, action);
+#endif
 }
 
+#ifdef XEN
+int request_irq_vector(unsigned int vector,
+		void (*handler)(int, void *, struct cpu_user_regs *),
+		unsigned long irqflags, const char * devname, void *dev_id)
+{
+	struct irqaction * action;
+	int retval;
+
+	/*
+	 * Sanity-check: shared interrupts must pass in a real dev-ID,
+	 * otherwise we'll have trouble later trying to figure out
+	 * which interrupt is which (messes up the interrupt freeing logic etc).
+	 *                          */
+	if (vector >= NR_VECTORS)
+		return -EINVAL;
+	if (!handler)
+		return -EINVAL;
+
+	action = xmalloc(struct irqaction);
+	if (!action)
+		return -ENOMEM;
+
+	action->handler = handler;
+	action->name = devname;
+	action->dev_id = dev_id;
+
+	retval = setup_vector(vector, action);
+	if (retval)
+		xfree(action);
+
+	return retval;
+}
+#endif
+
 void __init
 init_IRQ (void)
 {
