@@ -28,11 +28,11 @@
 #include <asm/hvm/support.h>
 #include <asm/current.h>
 
-#define domain_vrtc(d)   (&(d)->arch.hvm_domain.pl_time.vrtc)
-#define vcpu_vrtc(vcpu)  (domain_vrtc((vcpu)->domain))
-#define vrtc_domain(rtc) (container_of((rtc), struct domain, \
-                                       arch.hvm_domain.pl_time.vrtc))
-#define vrtc_vcpu(rtc)   (vrtc_domain(rtc)->vcpu[0])
+#define domain_vrtc(x) (&(x)->arch.hvm_domain.pl_time.vrtc)
+#define vcpu_vrtc(x)   (domain_vrtc((x)->domain))
+#define vrtc_domain(x) (container_of((x), struct domain, \
+                                     arch.hvm_domain.pl_time.vrtc))
+#define vrtc_vcpu(x)   (vrtc_domain(x)->vcpu[0])
 
 static void rtc_periodic_cb(struct vcpu *v, void *opaque)
 {
@@ -59,8 +59,8 @@ static void rtc_timer_update(RTCState *s)
 
         period = 1 << (period_code - 1); /* period in 32 Khz cycles */
         period = DIV_ROUND((period * 1000000000ULL), 32768); /* period in ns */
-        create_periodic_time(v, &s->pt, period, RTC_IRQ,
-                             0, rtc_periodic_cb, s);
+        create_periodic_time(v, &s->pt, period, period, RTC_IRQ,
+                             rtc_periodic_cb, s);
     }
     else
     {
@@ -160,6 +160,7 @@ static inline int from_bcd(RTCState *s, int a)
 static void rtc_set_time(RTCState *s)
 {
     struct tm *tm = &s->current_tm;
+    struct domain *d = vrtc_domain(s);
     unsigned long before, after; /* XXX s_time_t */
       
     ASSERT(spin_is_locked(&s->lock));
@@ -180,6 +181,12 @@ static void rtc_set_time(RTCState *s)
 
     after = mktime(tm->tm_year, tm->tm_mon, tm->tm_mday,
                    tm->tm_hour, tm->tm_min, tm->tm_sec);
+
+    /* We use the guest's setting of the RTC to define the local-time 
+     * offset for this domain. */
+    d->time_offset_seconds += (after - before);
+    update_domain_wallclock_time(d);
+    /* Also tell qemu-dm about it so it will be remembered for next boot. */
     send_timeoffset_req(after - before);
 }
 
