@@ -65,21 +65,20 @@ static atomic_t gdb_smp_paused_count;
 static void gdb_smp_pause(void);
 static void gdb_smp_resume(void);
 
-static char opt_gdb[30] = "none";
+static char opt_gdb[30];
 string_param("gdb", opt_gdb);
 
 static void gdbstub_console_puts(const char *str);
 
 /* value <-> char (de)serialzers */
-char
+static char
 hex2char(unsigned long x)
 {
     const char array[] = "0123456789abcdef";
-
     return array[x & 15];
 }
 
-int
+static unsigned int
 char2hex(unsigned char c)
 {
     if ( (c >= '0') && (c <= '9') )
@@ -93,13 +92,13 @@ char2hex(unsigned char c)
     return -1;
 }
 
-char
+static unsigned char
 str2hex(const char *str)
 {
     return (char2hex(str[0]) << 4) | char2hex(str[1]);
 }
 
-unsigned long
+static unsigned long
 str2ulong(const char *str, unsigned long bytes)
 {
     unsigned long x = 0;
@@ -116,6 +115,28 @@ str2ulong(const char *str, unsigned long bytes)
     return x;
 }
 
+static unsigned long
+str_to_native_ulong(const char *str)
+{
+    unsigned long x = 0, i = 0;
+
+    while ( *str && (i < BYTES_PER_LONG) )
+    {
+#ifdef __BIG_ENDIAN
+        x <<= 8;
+        x += str2hex(str);
+#elif defined(__LITTLE_ENDIAN)
+        x += (unsigned long)str2hex(str) << (i*8);
+#else
+# error unknown endian
+#endif
+        str += 2;
+        i++;
+    }
+
+    return x;
+}
+
 /* gdb io wrappers */
 static signed long
 gdb_io_write(const char *buf, unsigned long len, struct gdb_context *ctx)
@@ -488,7 +509,7 @@ process_command(struct cpu_user_regs *regs, struct gdb_context *ctx)
             return 0;
         }
         ptr++;
-        val = str2ulong(ptr, sizeof(unsigned long));
+        val = str_to_native_ulong(ptr);
         gdb_arch_write_reg(addr, val, regs, ctx);
         break;
     case 'D':
@@ -587,10 +608,6 @@ __trap_to_gdb(struct cpu_user_regs *regs, unsigned long cookie)
     watchdog_disable();
     console_start_sync();
 
-    /* Shouldn't really do this, but otherwise we stop for no
-       obvious reason, which is Bad */
-    printk("Waiting for GDB to attach...\n");
-
     gdb_arch_enter(regs);
     gdb_ctx->signum = gdb_arch_signal_num(regs, cookie);
 
@@ -625,10 +642,19 @@ __trap_to_gdb(struct cpu_user_regs *regs, unsigned long cookie)
 void __init
 initialise_gdb(void)
 {
+    if ( *opt_gdb == '\0' )
+        return;
+
     gdb_ctx->serhnd = serial_parse_handle(opt_gdb);
-    if ( gdb_ctx->serhnd != -1 )
-        printk("GDB stub initialised.\n");
+    if ( gdb_ctx->serhnd == -1 )
+    {
+        printk("Bad gdb= option '%s'\n", opt_gdb);
+        return;
+    }
+
     serial_start_sync(gdb_ctx->serhnd);
+
+    printk("GDB stub initialised.\n");
 }
 
 static void gdb_pause_this_cpu(void *unused)
